#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/init.h>
#include <linux/moduleparam.h>
#include <linux/rtnetlink.h>
#include <net/rtnetlink.h>
#include <net/arp.h>
#include <linux/u64_stats_sync.h>

#include "crypto0.h"

static struct net_device *crypto_dev = NULL;

struct crypto_priv
{
    struct net_device_stats priv_stats;
    struct net_device *slave;
};

rx_handler_result_t handle_frame(struct sk_buff **pskb)
{
    struct sk_buff *skb = *pskb;

    /* printk(KERN_WARNING "%s: Address src=%pM dst=%pM len=%d data_len=%d mac_len=%d\n", skb->dev->name, eth_hdr(skb)->h_source, eth_hdr(skb)->h_dest, skb->len, skb->data_len, skb->mac_len); */
#if 1
    /* We need to remove scb from slave device and push it to crypto device */
    if (crypto_dev)
    {
        struct crypto_priv *priv = netdev_priv(crypto_dev);
        
        priv->priv_stats.rx_packets++;
        priv->priv_stats.rx_bytes += skb->len;
        
        printk("injecting frame from %s to %s\n", skb->dev->name, crypto_dev->name);
        skb->dev = crypto_dev;
        /* netif_receive_skb(skb);
        return RX_HANDLER_CONSUMED; */
        return RX_HANDLER_ANOTHER;
    }
#endif
    return RX_HANDLER_PASS;
}

int crypto_open(struct net_device *dev)
{
    printk("%s: device opened\n", dev->name);
    netif_start_queue(dev);
    return 0;
}

int crypto_close(struct net_device *dev)
{
    printk("%s: device closed\n", dev->name);
    netif_stop_queue(dev);
    return 0;
}

netdev_tx_t crypto_xmit(struct sk_buff *skb, struct net_device *dev)
{
    struct crypto_priv *priv = netdev_priv(dev);

    printk("%s: crypto_xmit\n", dev->name);

    priv->priv_stats.tx_packets++;
    priv->priv_stats.tx_bytes += skb->len;

    /* We need to crypt the frame and push in to the slave device */
#if 1
    if (priv->slave)
    {
        skb->dev = priv->slave;
        skb->priority = 1;
        printk("injecting frame from %s to %s\n", dev->name, skb->dev->name);
        dev_queue_xmit (skb);
        return 0;
    }
#else
    kfree_skb(skb);
#endif

    return NETDEV_TX_OK;
}

int crypto_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    int err;
    struct net_device *slave;
    struct crypto_priv *priv = netdev_priv(dev);
    struct crypto_userinfo info;
    struct crypto_userinfo *uptr = (struct crypto_userinfo *)ifr->ifr_data;

    /* only authorized users can control the interface */
    if (cmd == SIOCCRYPTOSETINFO && !capable(CAP_NET_ADMIN))
        return -EPERM;

    /* process the command */
    switch(cmd) {
        case SIOCCRYPTOGETINFO: /* return configuration to user space */

            memset(info.name, 0, CRYPTO_NAMELEN);
            if (priv->slave)
                strncpy(info.name, priv->slave->name, CRYPTO_NAMELEN-1);

            /* return the data structure to  user space */
            err = copy_to_user(uptr, &info, sizeof(info));
            if (err) return err;
            break;

        case SIOCCRYPTOSETINFO:
            /* retrieve the data structure from user space */
            err = copy_from_user(&info, uptr, sizeof(info));
            if (err) return err;

            /* interface name */
            slave = __dev_get_by_name(&init_net, info.name);
            if (!slave)
                return -ENODEV;

            if (slave->type != ARPHRD_ETHER && slave->type != ARPHRD_LOOPBACK)
                return -EINVAL;

            /* The interface is good, get hold of it */
            priv->slave = slave;

            /* also, and clone its IP, MAC and other information */
            memcpy(dev->dev_addr,  slave->dev_addr,  sizeof(slave->dev_addr));
            memcpy(dev->broadcast, slave->broadcast, sizeof(slave->broadcast));


	    netdev_rx_handler_register(slave, handle_frame, NULL);
	    printk(KERN_WARNING "register rx handler for %s", slave->name);

            break;

        default:
            return -EOPNOTSUPP;
    }

    return 0;
}

struct net_device_stats *crypto_get_stats(struct net_device *dev)
{
    return &((struct crypto_priv *)netdev_priv(dev))->priv_stats;
}

static struct net_device_ops crypto_net_device_ops =
{
    .ndo_open = crypto_open,
    .ndo_stop = crypto_close,
    .ndo_get_stats = crypto_get_stats,
    .ndo_start_xmit = crypto_xmit,
    .ndo_do_ioctl = crypto_ioctl,
};

void crypto_setup(struct net_device *dev)
{
    ether_setup(dev);

    memset(netdev_priv(dev), 0, sizeof(struct crypto_priv));

    dev->netdev_ops = &crypto_net_device_ops;

    random_ether_addr(dev->dev_addr);
}

int __init crypto_init(void)
{
    int err = 0;

    crypto_dev = alloc_netdev(sizeof(struct crypto_priv), "crypto", crypto_setup);
    if (crypto_dev == NULL) 
        return -ENOMEM;

    if (( err = dev_alloc_name(crypto_dev, crypto_dev->name) )) {
        printk(KERN_WARNING "%s: allocate name, error %i\n", crypto_dev->name, err);
        return -EIO;
    }

    register_netdev(crypto_dev);

    printk("loaded\n");

    return 0;
}

void __exit crypto_exit(void)
{
    struct crypto_priv *priv;

    if (crypto_dev)
    {
	priv = netdev_priv(crypto_dev);
        if (priv->slave)
        {
	    printk(KERN_WARNING "unregister rx handler for %s\n", priv->slave->name);
	    rtnl_lock();
	    netdev_rx_handler_unregister(priv->slave);
	    rtnl_unlock();
        }
        unregister_netdev(crypto_dev);
        free_netdev(crypto_dev);
    }

    printk("unloaded\n");
}

module_init(crypto_init);
module_exit(crypto_exit);
MODULE_LICENSE("GPL");
